﻿using System;
using SDX.Compiler;
using Mono.Cecil;
using Mono.Cecil.Cil;
using System.Linq;
using System.Collections.Generic;

public class WildernessGameStage : IPatcherMod
{
    public enum GameStagePartyPatchState
    {
        START,
        FINDINITPOS,
        FINDCALCPOS,
        FINDINSERTINCPOS,
        FINDINJUMP,
    }

    public bool Link(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        InsertCallSetTownship(gameModule, modModule);
        InsertTownshipChunkMethod(gameModule, modModule);
        SetSpawnManagerBiomesNZMethod(gameModule, modModule);
        InsertGameStageDefinitionRead(gameModule, modModule);
        InsertGetWanderingHordeGameStage(gameModule, modModule);
        InsertInitWildernessGameStageDic(gameModule, modModule);
        InsertUnloadChunk(gameModule, modModule);
        AddLockAllEnemySpawningUntilArg(gameModule, modModule);
        InsertWriteHGAbstructHubDataCall(gameModule, modModule);
        InsertHubCellRead(gameModule, modModule);
        InsertInitHGAbstructHubDataCall(gameModule, modModule);
        InsertSetAreaHubDataLotCall(gameModule, modModule);
        return true;
    }
    public void InsertSetAreaHubDataLotCall(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "HGSockets" && d.Namespace == "RWG2.HubGenerators");


        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "GenerateLots");


        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos1=null;
        Instruction insertPos2 = null;
        VariableDefinition vd1 = null;
        VariableDefinition vd2 = null;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).FullName == "System.Void System.Collections.Generic.List`1<RWG2.HubCell/Lot>::Add(!0)")
            {
                if (inst.Previous.OpCode == OpCodes.Ldloc_S && inst.Previous.Operand.GetType() == typeof(VariableDefinition))
                {
                    VariableDefinition vd = (VariableDefinition)inst.Previous.Operand;
                    if (vd.Index == 29)
                    {
                        insertPos1 = inst.Next;
                        vd1 = vd;
                    }
                    else
                    {
                        insertPos2 = inst.Next;
                        vd2 = vd;

                    }

                }
            }
        }

        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");
        MethodDefinition setAreaHubDataLot = wildernessGameStageMethod.Methods.First(d => d.Name == "SetAreaHubDataLot");
        MethodReference setAreaHubDataLotRef = gameModule.Import(setAreaHubDataLot);

        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Ldloc_S, vd1));
        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Call, setAreaHubDataLotRef));

        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Ldloc_S, vd2));
        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Call, setAreaHubDataLotRef));




    }
    public void RemoveHGSotcketsCZCall(ModuleDefinition gameModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "HGSockets");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "GenerateLots");

        Instruction removeCall = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).Name == "CZ")
            {
                removeCall = inst;

            }
        }


        ILProcessor ilLockAllEnemySpawningUntil = methodDef.Body.GetILProcessor();
        ilLockAllEnemySpawningUntil.Remove(removeCall.Previous);//inst IL_0d06: ldloc.0 opcode ldloc.0 operand null
        ilLockAllEnemySpawningUntil.Remove(removeCall.Previous);//inst IL_0d05: ldarg.0 opcode ldarg.0 operand null
        ilLockAllEnemySpawningUntil.Remove(removeCall);//inst IL_0d07: call System.Void RWG2.HubGenerators.HGSockets::CZ(System.Single) opcode call operand System.Void RWG2.HubGenerators.HGSockets::CZ(System.Single)

    }
    public void AddLockAllEnemySpawningUntilArg(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "ChunkAreaBiomeSpawnData");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "LockAllEnemySpawningUntil");
        TypeDefinition world = gameModule.Types.First(d => d.Name == "World");


        ParameterDefinition pd = new ParameterDefinition(world);
        methodDef.Parameters.Add(pd);
        CallAddLockAllEnemySpawningUntilArgWorld(gameModule, modModule);

        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");

        MethodDefinition getWildernessGameStageGroup = wildernessGameStageMethod.Methods.First(d => d.Name == "GetWildernessGameStageGroup");
        MethodReference getWildernessGameStageGroupRef = gameModule.Import(getWildernessGameStageGroup);

        Instruction insertPos = null;
        foreach(Instruction inst in methodDef.Body.Instructions)
        {
            if(inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).Name == "IsEnemyGroup")
            {
                if (inst.Previous.OpCode == OpCodes.Ldfld && inst.Previous.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Previous.Operand).Name == "entityGroupRefName")
                {
                    insertPos = inst;
                }

            }
        }
        ILProcessor ilLockAllEnemySpawningUntil = methodDef.Body.GetILProcessor();
        ilLockAllEnemySpawningUntil.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        ilLockAllEnemySpawningUntil.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_3));
        ilLockAllEnemySpawningUntil.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, getWildernessGameStageGroupRef));

    }
    public void CallAddLockAllEnemySpawningUntilArgWorld(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "World");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "LockAreaMasterChunksAround");
        Instruction insertPos = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if(inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).Name == "LockAllEnemySpawningUntil")
            {
                insertPos = inst;
            }
        }
        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));

        methodDef = classDef.Methods.First(d => d.Name == "OnUpdateTick");
        insertPos = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).Name == "LockAllEnemySpawningUntil")
            {
                insertPos = inst;
            }
        }
        proc1 = methodDef.Body.GetILProcessor();
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));

    }

    public void InsertUnloadChunk(ModuleDefinition gameModule, ModuleDefinition modModule)
    {

        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "Chunk");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "OnUnload");

        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");
        MethodDefinition unLoadChunk = wildernessGameStageMethod.Methods.First(d => d.Name == "UnLoadChunk");
        MethodReference unLoadChunkRef = gameModule.Import(unLoadChunk);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos = methodDef.Body.Instructions[0];
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, unLoadChunkRef));

    }

    public void InsertWriteHGAbstructHubDataCall(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "HGAbstract" && d.Namespace == "RWG2.HubGenerators");

        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Write");

        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");
        MethodDefinition writeHubChunkFlgData = wildernessGameStageMethod.Methods.First(d => d.Name == "WriteHubChunkFlgData");
        MethodReference writeHubChunkFlgDataRef = gameModule.Import(writeHubChunkFlgData);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos = methodDef.Body.Instructions[methodDef.Body.Instructions.Count - 1];
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, writeHubChunkFlgDataRef));

    }
    public void InsertHubCellRead(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "HubCell" && d.Namespace == "RWG2");

        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Read");
        FieldDefinition hubGeneratorField = classDef.Fields.First(d => d.Name == "HubGenerator");

        Instruction insertPos = null;
        foreach(Instruction inst in methodDef.Body.Instructions)
        {
            if(inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "System.Void RWG2.HubGenerators.HGAbstract::Read(System.IO.BinaryReader,System.Int32)")
            {
                insertPos = inst.Next;
            }
        }


        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");
        MethodDefinition readHubChunkFlgData = wildernessGameStageMethod.Methods.First(d => d.Name == "ReadHubChunkFlgData");
        MethodReference readHubChunkFlgDataRef = gameModule.Import(readHubChunkFlgData);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction newJumpTo = Instruction.Create(OpCodes.Ldarg_0);
        proc1.InsertBefore(insertPos, newJumpTo);
        
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, readHubChunkFlgDataRef));
    }
    public void InsertInitHGAbstructHubDataCall(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "HGSockets" && d.Namespace == "RWG2.HubGenerators");

        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Init");

        TypeDefinition hgAbstractDef = gameModule.Types.First(d => d.Name == "HGAbstract" && d.Namespace == "RWG2.HubGenerators");
        
        FieldDefinition hubCellFld = hgAbstractDef.Fields.First(d => d.Name == "hubCell");



        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");
        MethodDefinition initHGAbstructHubData = wildernessGameStageMethod.Methods.First(d => d.Name == "InitHGAbstructHubData");
        MethodReference initHGAbstructHubDataRef = gameModule.Import(initHGAbstructHubData);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos = null;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if(inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "System.Void RWG2.HubGenerators.HGSockets::DZ()")
            {
                insertPos = inst.Next;
            }
        }
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldfld, hubCellFld));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, initHGAbstructHubDataRef));

    }

    
    public void InsertInitWildernessGameStageDic(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "ChunkManager");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Init");

        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");
        MethodDefinition initWildernessGameStageDic = wildernessGameStageMethod.Methods.First(d => d.Name == "InitWildernessGameStageDic");
        MethodReference initWildernessGameStageDicRef = gameModule.Import(initWildernessGameStageDic);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        proc1.InsertBefore(methodDef.Body.Instructions[0], Instruction.Create(OpCodes.Call, initWildernessGameStageDicRef));

    }
    public void InsertGetWanderingHordeGameStage(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "AIDirectorGameStagePartySpawner");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "ResetPartyLevel");

        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");
        MethodDefinition getWanderingHordeGameStage = wildernessGameStageMethod.Methods.First(d => d.Name == "GetWanderingHordeGameStage");

        MethodReference getWanderingHordeGameStageRef = gameModule.Import(getWanderingHordeGameStage);


        Instruction insertPos = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Rem)
            {
                insertPos = inst;
            }
        }
        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, getWanderingHordeGameStageRef));
        proc1.Remove(insertPos);

    }
    public void InsertGameStageDefinitionRead(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "GameStagesFromXml");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Load");

        Instruction insertPosition = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {

            if (inst.OpCode == OpCodes.Stsfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Single GameStageDefinition::DaysAliveDifficultyBonus")
            {
                insertPosition = inst.Next;

            }
        }
        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Brfalse_S && inst.Operand.GetType() == typeof(Instruction) && inst.Operand == insertPosition)
            {
                jumpList.Add(inst);
            }
        }

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction newJumpTo = Instruction.Create(OpCodes.Ldloc_S, proc1.Body.Variables[4]);
        proc1.InsertBefore(insertPosition, newJumpTo);//V_4 is XmlElement

        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");
        MethodDefinition setGamestageUpdateWaitTime = wildernessGameStageMethod.Methods.First(d => d.Name == "SetGamestageUpdateWaitTime");
        

        if (setGamestageUpdateWaitTime == null)
        {
            Console.WriteLine("WildernessGameStageMethod SetGamestageUpdateWaitTime == null");
        }
        Console.WriteLine("WildernessGameStageMethod SetGamestageUpdateWaitTime find");
        MethodReference setGamestageUpdateWaitTimeRef = gameModule.Import(setGamestageUpdateWaitTime);
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, setGamestageUpdateWaitTimeRef));

        MethodDefinition setWanderingHordeStageMax = wildernessGameStageMethod.Methods.First(d => d.Name == "SetWanderingHordeStageMax");
        MethodReference setWanderingHordeStageMaxRef = gameModule.Import(setWanderingHordeStageMax);
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Ldloc_S, proc1.Body.Variables[4]));
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, setWanderingHordeStageMaxRef));


        foreach (Instruction jump in jumpList)
        {
            jump.Operand = newJumpTo;
        }
    }

    public void SetSpawnManagerBiomesNZMethod(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition spawnManagerBiomes = gameModule.Types.First(d => d.Name == "SpawnManagerBiomes");
        MethodDefinition spawnManagerBiomesNZ = spawnManagerBiomes.Methods.First(d => d.Name == "NZ");



        TypeDefinition spawnManagerAbstract = gameModule.Types.First(d => d.Name == "SpawnManagerAbstract");
        
        FieldDefinition spawnManagerAbstractworld = spawnManagerAbstract.Fields.First(d => d.Name == "world");

        Instruction groupListGetPos = null;
        Instruction isEnemyGroupPos = null;
        Instruction getRandomFromGroup = null;
        Instruction getRandomFromGroupBeforeJump = null;

        foreach (Instruction inst in spawnManagerBiomesNZ.Body.Instructions)
        {

            if (inst.OpCode == OpCodes.Ldsfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "list")
            {
                groupListGetPos = inst;
            }
            if (inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).Name == "IsEnemyGroup")
            {
                isEnemyGroupPos = inst;
            }
            if (inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).Name == "GetRandomFromGroup")
            {
                getRandomFromGroup = inst;
            }
            if (inst.OpCode == OpCodes.Bne_Un_S && inst.Operand.GetType() == typeof(Instruction) && ((Instruction)inst.Operand).OpCode == OpCodes.Ldloc_S && ((Instruction)inst.Operand).Operand.GetType() == typeof(VariableDefinition) && ((VariableDefinition)((Instruction)inst.Operand).Operand).Index == 8)
            {
                getRandomFromGroupBeforeJump = inst;//inst IL_03b2: bne.un.s IL_03bf opcode bne.un.s operand IL_03bf: ldloc.s V_8
            }

        }
        ILProcessor ilspawnManagerBiomesNZ = spawnManagerBiomesNZ.Body.GetILProcessor();

        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");
        MethodDefinition getBiomeSpawnEntityGroupList = wildernessGameStageMethod.Methods.First(d => d.Name == "GetBiomeSpawnEntityGroupList");
        MethodReference getBiomeSpawnEntityGroupListRef = gameModule.Import(getBiomeSpawnEntityGroupList);

        MethodDefinition getWildernessGameStageGroup = wildernessGameStageMethod.Methods.First(d => d.Name == "GetWildernessGameStageGroup");
        MethodReference getWildernessGameStageGroupRef = gameModule.Import(getWildernessGameStageGroup);


        ilspawnManagerBiomesNZ.InsertBefore(groupListGetPos, Instruction.Create(OpCodes.Ldarg_3));
        ilspawnManagerBiomesNZ.InsertBefore(groupListGetPos, Instruction.Create(OpCodes.Ldarg_0));
        ilspawnManagerBiomesNZ.InsertBefore(groupListGetPos, Instruction.Create(OpCodes.Ldfld, spawnManagerAbstractworld));
        ilspawnManagerBiomesNZ.InsertBefore(groupListGetPos, Instruction.Create(OpCodes.Call, getBiomeSpawnEntityGroupListRef));

        ilspawnManagerBiomesNZ.Remove(groupListGetPos.Next);//inst IL_01cd: ldloc.s V_9 opcode ldloc.s operand V_9
        ilspawnManagerBiomesNZ.Remove(groupListGetPos.Next);//inst IL_01cf: ldfld System.String BiomeDefinition::m_sBiomeName opcode ldfld operand System.String BiomeDefinition::m_sBiomeName
        ilspawnManagerBiomesNZ.Remove(groupListGetPos.Next);//inst IL_01d4: callvirt T2 DictionarySave`2<System.String,BiomeSpawnEntityGroupList>::get_Item(T1) opcode callvirt operand T2 DictionarySave`2<System.String,BiomeSpawnEntityGroupList>::get_Item(T1)
        ilspawnManagerBiomesNZ.Remove(groupListGetPos);//inst IL_01c8: ldsfld DictionarySave`2<System.String,BiomeSpawnEntityGroupList> BiomeSpawningClass::list opcode ldsfld operand DictionarySave`2<System.String,BiomeSpawnEntityGroupList> BiomeSpawningClass::list

        VariableDefinition vd = new VariableDefinition("V_22", gameModule.Import(typeof(string)));


        ilspawnManagerBiomesNZ.Body.Variables.Add(vd);

        ilspawnManagerBiomesNZ.InsertBefore(isEnemyGroupPos, Instruction.Create(OpCodes.Ldarg_3));
        ilspawnManagerBiomesNZ.InsertBefore(isEnemyGroupPos, Instruction.Create(OpCodes.Ldarg_0));
        ilspawnManagerBiomesNZ.InsertBefore(isEnemyGroupPos, Instruction.Create(OpCodes.Ldfld, spawnManagerAbstractworld));
        ilspawnManagerBiomesNZ.InsertBefore(isEnemyGroupPos, Instruction.Create(OpCodes.Call, getWildernessGameStageGroupRef));
        ilspawnManagerBiomesNZ.InsertBefore(isEnemyGroupPos, Instruction.Create(OpCodes.Stloc_S, vd));
        ilspawnManagerBiomesNZ.InsertBefore(isEnemyGroupPos, Instruction.Create(OpCodes.Ldloc_S, vd));

        ilspawnManagerBiomesNZ.InsertBefore(getRandomFromGroup, Instruction.Create(OpCodes.Ldarg_3));
        ilspawnManagerBiomesNZ.InsertBefore(getRandomFromGroup, Instruction.Create(OpCodes.Ldarg_0));
        ilspawnManagerBiomesNZ.InsertBefore(getRandomFromGroup, Instruction.Create(OpCodes.Ldfld, spawnManagerAbstractworld));
        ilspawnManagerBiomesNZ.InsertBefore(getRandomFromGroup, Instruction.Create(OpCodes.Call, getWildernessGameStageGroupRef));
        ilspawnManagerBiomesNZ.InsertBefore(getRandomFromGroup, Instruction.Create(OpCodes.Stloc_S, vd));
        ilspawnManagerBiomesNZ.InsertBefore(getRandomFromGroup, Instruction.Create(OpCodes.Ldloc_S, vd));

    }

    public void InsertTownshipChunkMethod(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        
        TypeDefinition chunk = gameModule.Types.First(d => d.Name == "Chunk");
        MethodDefinition chunkctor = chunk.Methods.First(d => d.Name == ".ctor");
        FieldDefinition chunkTownShipJA = chunk.Fields.First(d => d.Name == "TownShipJA");
        FieldDefinition chunkDominantTownShip = chunk.Fields.First(d => d.Name == "DominantTownShip");

        

        Instruction ctorInsertPos = null;
        Instruction ctorAreaMasterInsertPos = null;
        foreach (Instruction inst in chunkctor.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "JA")
            {
                ctorInsertPos = inst.Next;
            }
            if(inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "AreaMasterDominantBiome")
            {
                ctorAreaMasterInsertPos = inst.Next;
            }
        }
        ILProcessor ilchunkctor = chunkctor.Body.GetILProcessor();
        ilchunkctor.InsertBefore(ctorInsertPos, Instruction.Create(OpCodes.Ldarg_0));
        ilchunkctor.InsertBefore(ctorInsertPos, Instruction.Create(OpCodes.Ldc_I4,256));
        TypeReference byteRef = gameModule.Import(typeof(Byte));
        ilchunkctor.InsertBefore(ctorInsertPos, Instruction.Create(OpCodes.Newarr, byteRef));
        ilchunkctor.InsertBefore(ctorInsertPos, Instruction.Create(OpCodes.Stfld, chunkTownShipJA));

        MethodDefinition chunkYVQ = chunk.Methods.First(d => d.Name == "YVQ");
        Instruction yvqInsertPos = null;
        Instruction yvqAreaMasterInsertPos = null;
        Instruction yvqTownShipInsertPos = null;
        foreach (Instruction inst in chunkYVQ.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Ldfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "JA")
            {
                yvqInsertPos = inst.Next.Next.Next.Next.Next.Next.Next;
            }
            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "AreaMasterDominantBiome")
            {
                yvqAreaMasterInsertPos = inst.Next;

            }
            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "DominantBiome")
            {
                yvqTownShipInsertPos = inst.Next;

            }
        }

        ILProcessor ilchunkYVQ = chunkYVQ.Body.GetILProcessor();
        MethodReference binaryReaderReadRef = (MethodReference)yvqInsertPos.Previous.Previous.Previous.Previous.Operand;
        ilchunkYVQ.InsertBefore(yvqInsertPos, Instruction.Create(OpCodes.Ldarg_1));
        ilchunkYVQ.InsertBefore(yvqInsertPos, Instruction.Create(OpCodes.Ldarg_0));
        ilchunkYVQ.InsertBefore(yvqInsertPos, Instruction.Create(OpCodes.Ldfld, chunkTownShipJA));
        ilchunkYVQ.InsertBefore(yvqInsertPos, Instruction.Create(OpCodes.Ldc_I4_0));
        ilchunkYVQ.InsertBefore(yvqInsertPos, Instruction.Create(OpCodes.Ldc_I4, 256));
        ilchunkYVQ.InsertBefore(yvqInsertPos, Instruction.Create(OpCodes.Callvirt, binaryReaderReadRef));
        ilchunkYVQ.InsertBefore(yvqInsertPos, Instruction.Create(OpCodes.Dup));
        ilchunkYVQ.InsertBefore(yvqInsertPos, Instruction.Create(OpCodes.Pop));
        ilchunkYVQ.InsertBefore(yvqInsertPos, Instruction.Create(OpCodes.Pop));

        MethodReference binaryReaderReadByteRef = (MethodReference)yvqTownShipInsertPos.Previous.Previous.Previous.Previous.Operand;
        ilchunkYVQ.InsertBefore(yvqTownShipInsertPos, Instruction.Create(OpCodes.Ldarg_0));
        ilchunkYVQ.InsertBefore(yvqTownShipInsertPos, Instruction.Create(OpCodes.Ldarg_1));
        ilchunkYVQ.InsertBefore(yvqTownShipInsertPos, Instruction.Create(OpCodes.Callvirt, binaryReaderReadByteRef));
        ilchunkYVQ.InsertBefore(yvqTownShipInsertPos, Instruction.Create(OpCodes.Stfld, chunkDominantTownShip));

        MethodDefinition chunkNVQ = chunk.Methods.First(d => d.Name == "NVQ");
        Instruction nvqInsertPos = null;
        Instruction nvqAreaMasterInsertPos = null;
        Instruction nvqTownShipInsertPos = null;
        foreach (Instruction inst in chunkNVQ.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Ldfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "JA")
            {
                nvqInsertPos = inst.Next.Next;
            }
            if (inst.OpCode == OpCodes.Ldfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "DominantBiome")
            {
                nvqTownShipInsertPos = inst.Next.Next;
            }
            if (inst.OpCode == OpCodes.Ldfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "AreaMasterDominantBiome")
            {
                nvqAreaMasterInsertPos = inst.Next.Next;
            }
        }
        ILProcessor ilchunkNVQ = chunkNVQ.Body.GetILProcessor();
        MethodReference binarWriterWriteRef = (MethodReference)nvqInsertPos.Previous.Operand;
        ilchunkNVQ.InsertBefore(nvqInsertPos, Instruction.Create(OpCodes.Ldarg_1));
        ilchunkNVQ.InsertBefore(nvqInsertPos, Instruction.Create(OpCodes.Ldarg_0));
        ilchunkNVQ.InsertBefore(nvqInsertPos, Instruction.Create(OpCodes.Ldfld, chunkTownShipJA));
        ilchunkNVQ.InsertBefore(nvqInsertPos, Instruction.Create(OpCodes.Callvirt, binarWriterWriteRef));

        MethodReference binarWriterWriteByteRef = (MethodReference)nvqTownShipInsertPos.Previous.Operand;
        ilchunkNVQ.InsertBefore(nvqTownShipInsertPos, Instruction.Create(OpCodes.Ldarg_1));
        ilchunkNVQ.InsertBefore(nvqTownShipInsertPos, Instruction.Create(OpCodes.Ldarg_0));
        ilchunkNVQ.InsertBefore(nvqTownShipInsertPos, Instruction.Create(OpCodes.Ldfld, chunkDominantTownShip));
        ilchunkNVQ.InsertBefore(nvqTownShipInsertPos, Instruction.Create(OpCodes.Callvirt, binarWriterWriteByteRef));

        MethodDefinition chunkReset = chunk.Methods.First(d => d.Name == "Reset");
        Instruction resetInsertPos = null;
        foreach (Instruction inst in chunkReset.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "AreaMasterDominantBiome")
            {
                resetInsertPos = inst.Next;
            }
        }
        ILProcessor ilchunkReset = chunkReset.Body.GetILProcessor();


        ilchunkReset.InsertBefore(resetInsertPos, Instruction.Create(OpCodes.Ldarg_0));
        ilchunkReset.InsertBefore(resetInsertPos, Instruction.Create(OpCodes.Ldc_I4_0));
        ilchunkReset.InsertBefore(resetInsertPos, Instruction.Create(OpCodes.Stfld, chunkDominantTownShip));

        TypeDefinition chunkCluster = gameModule.Types.First(d => d.Name == "ChunkCluster");
        MethodDefinition lightChunk = chunkCluster.Methods.First(d => d.Name == "LightChunk");
        Instruction lightChunkInsertPos = null;
        foreach (Instruction inst in lightChunk.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).Name == "CalcDominantBiome")
            {
                lightChunkInsertPos = inst.Next;
            }
        }
        ILProcessor illightChunk = lightChunk.Body.GetILProcessor();


        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");
        MethodDefinition calcDominantTownShip = wildernessGameStageMethod.Methods.First(d => d.Name == "CalcDominantTownShip");
        MethodReference calcDominantTownShipRef = gameModule.Import(calcDominantTownShip);

        illightChunk.InsertBefore(lightChunkInsertPos, Instruction.Create(OpCodes.Ldarg_1));
        illightChunk.InsertBefore(lightChunkInsertPos, Instruction.Create(OpCodes.Call, calcDominantTownShipRef));


        MethodDefinition isAreaMasterDBiomeI = chunk.Methods.First(d => d.Name == "IsAreaMasterDominantBiomeInitialized");

        Instruction isAreaMasterDBiomeIInsertPos = isAreaMasterDBiomeI.Body.Instructions[isAreaMasterDBiomeI.Body.Instructions.Count - 2];

        ILProcessor ilisAreaMasterDBiomeI = isAreaMasterDBiomeI.Body.GetILProcessor();

        ilisAreaMasterDBiomeI.InsertBefore(isAreaMasterDBiomeIInsertPos, Instruction.Create(OpCodes.Ldarg_0));
        ilisAreaMasterDBiomeI.InsertBefore(isAreaMasterDBiomeIInsertPos, Instruction.Create(OpCodes.Call, calcDominantTownShipRef));


        TypeDefinition rwgTerrainGenerator = gameModule.Types.First(d => d.Name == "RWGTerrainGenerator");
        MethodDefinition generateTerrain = rwgTerrainGenerator.Methods.First(d => d.Name == "GenerateTerrain");

        Instruction generateTerrainInsertPos = null;
        FieldDefinition wzHGAbstructRef = null;
        foreach (Instruction inst in generateTerrain.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).Name == "SetBiomeId")
            {
                generateTerrainInsertPos = inst.Next;
            }
            if (inst.OpCode == OpCodes.Ldfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "WZ")
            {
                wzHGAbstructRef = (FieldDefinition)inst.Operand;
            }
        }
        ILProcessor ilgenerateTerrain = generateTerrain.Body.GetILProcessor();

        MethodDefinition checkAndSetTownShipID = wildernessGameStageMethod.Methods.First(d => d.Name == "CheckAndSetTownShipID");
        MethodReference checkAndSetTownShipIDpRef = gameModule.Import(checkAndSetTownShipID);
        ilgenerateTerrain.InsertBefore(generateTerrainInsertPos, Instruction.Create(OpCodes.Ldloc_3));
        ilgenerateTerrain.InsertBefore(generateTerrainInsertPos, Instruction.Create(OpCodes.Ldloc_S, ilgenerateTerrain.Body.Variables[4]));
        ilgenerateTerrain.InsertBefore(generateTerrainInsertPos, Instruction.Create(OpCodes.Ldloc_1));
        ilgenerateTerrain.InsertBefore(generateTerrainInsertPos, Instruction.Create(OpCodes.Ldloc_2));

        ilgenerateTerrain.InsertBefore(generateTerrainInsertPos, Instruction.Create(OpCodes.Ldarg_0));
        ilgenerateTerrain.InsertBefore(generateTerrainInsertPos, Instruction.Create(OpCodes.Ldfld, wzHGAbstructRef));

        ilgenerateTerrain.InsertBefore(generateTerrainInsertPos, Instruction.Create(OpCodes.Ldarg_2));
        ilgenerateTerrain.InsertBefore(generateTerrainInsertPos, Instruction.Create(OpCodes.Call, checkAndSetTownShipIDpRef));

        MethodDefinition getChunkBiomeSpawnData = chunk.Methods.First(d => d.Name == "GetChunkBiomeSpawnData");
        FieldDefinition chunkLF = chunk.Fields.First(d => d.Name == "LF");

        TypeDefinition chunkAreaBiomeSpawnData = gameModule.Types.First(d => d.Name == "ChunkAreaBiomeSpawnData");
        FieldDefinition chunkAreaBiomeSpawnDataTownShip = chunkAreaBiomeSpawnData.Fields.First(d => d.Name == "TownShip");

        Instruction getChunkBiomeSpawnDataInsertPos = null;
        foreach (Instruction inst in getChunkBiomeSpawnData.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "LF")
            {

                getChunkBiomeSpawnDataInsertPos = inst.Next;
            }
        }
        ILProcessor ilgetChunkBiomeSpawnData = getChunkBiomeSpawnData.Body.GetILProcessor();

        Instruction newJumpTo = Instruction.Create(OpCodes.Ldarg_0);
        ilgetChunkBiomeSpawnData.InsertBefore(getChunkBiomeSpawnDataInsertPos, newJumpTo);
        ilgetChunkBiomeSpawnData.InsertBefore(getChunkBiomeSpawnDataInsertPos, Instruction.Create(OpCodes.Ldfld, chunkLF));        
        ilgetChunkBiomeSpawnData.InsertBefore(getChunkBiomeSpawnDataInsertPos, Instruction.Create(OpCodes.Ldarg_0));
        ilgetChunkBiomeSpawnData.InsertBefore(getChunkBiomeSpawnDataInsertPos, Instruction.Create(OpCodes.Ldfld, chunkDominantTownShip));
        ilgetChunkBiomeSpawnData.InsertBefore(getChunkBiomeSpawnDataInsertPos, Instruction.Create(OpCodes.Stfld, chunkAreaBiomeSpawnDataTownShip));

    }
    public void InsertCallSetTownship(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition biomeSpawningFromXml = gameModule.Types.First(d => d.Name == "BiomeSpawningFromXml");
        MethodDefinition biomeSpawningFromXmlLoad = biomeSpawningFromXml.Methods.First(d => d.Name == "Load");
        Instruction insertPos = null;
        Instruction insertCheckPos = null;
        foreach (Instruction inst in biomeSpawningFromXmlLoad.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Stloc_S && inst.Operand.GetType() == typeof(VariableDefinition) && ((VariableDefinition)inst.Operand).Index == 13)
            {
                insertPos = inst.Next;
            }
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).Name == "ContainsKey")
            {
                insertCheckPos = inst;
            }
        }
        ILProcessor ilProc = biomeSpawningFromXmlLoad.Body.GetILProcessor();
        ilProc.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldloc_S, biomeSpawningFromXmlLoad.Body.Variables[13]));
        ilProc.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldloc_S, biomeSpawningFromXmlLoad.Body.Variables[8]));

        TypeDefinition wildernessGameStageMethod = modModule.Types.First(d => d.Name == "WildernessGameStageMethod");
        MethodDefinition setTownship = wildernessGameStageMethod.Methods.First(d => d.Name == "SetTownship");
        MethodReference setTownshipRef = gameModule.Import(setTownship);

        ilProc.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, setTownshipRef));

        ilProc.Remove(insertCheckPos.Next);//inst IL_02f0: dup opcode dup operand null
        ilProc.Remove(insertCheckPos.Next);//inst IL_02f1: pop opcode pop operand null
        insertCheckPos.Next.OpCode = OpCodes.Br_S;
        ilProc.Remove(insertCheckPos.Previous);//inst IL_02e6: ldfld System.String BiomeSpawnEntityGroupData::entityGroupRefName opcode ldfld operand System.String BiomeSpawnEntityGroupData::entityGroupRefName
        ilProc.Remove(insertCheckPos.Previous);//inst IL_02e4: ldloc.s V_13 opcode ldloc.s operand V_13
        ilProc.Remove(insertCheckPos.Previous);//inst IL_02df: ldsfld DictionarySave`2<System.String,System.Collections.Generic.List`1<SEntityClassAndProb>> EntityGroups::list opcode ldsfld operand DictionarySave`2<System.String,System.Collections.Generic.List`1<SEntityClassAndProb>> EntityGroups::list
        ilProc.Remove(insertCheckPos);//inst IL_02eb: callvirt System.Boolean DictionarySave`2<System.String,System.Collections.Generic.List`1<SEntityClassAndProb>>::ContainsKey(T1) opcode callvirt operand System.Boolean DictionarySave`2<System.String,System.Collections.Generic.List`1<SEntityClassAndProb>>::ContainsKey(T1)
    }

    public void AddBiomeSpawnEntityGroupDataField(ModuleDefinition module)
    {
        TypeDefinition biomeSpawnEntityGroupData = module.Types.First(d => d.Name == "BiomeSpawnEntityGroupData");
        TypeReference tr = module.Import(typeof(List<byte>));
        AddFieldDefinition(biomeSpawnEntityGroupData, "TownShip", FieldAttributes.Public, tr);

    }
    public void AddChunkTownShipJAField(ModuleDefinition module)
    {
        TypeDefinition biomeSpawnEntityGroupData = module.Types.First(d => d.Name == "Chunk");
        TypeReference tr = module.Import(typeof(byte[]));
        AddFieldDefinition(biomeSpawnEntityGroupData, "TownShipJA", FieldAttributes.Public, tr);

    }
    public void AddChunkDominantTownShipField(ModuleDefinition module)
    {
        TypeDefinition chunk = module.Types.First(d => d.Name == "Chunk");
        TypeReference tr = module.Import(typeof(byte));
        AddFieldDefinition(chunk, "DominantTownShip", FieldAttributes.Public, tr);
    }
    public void AddChunkAreaBiomeSpawnDataTownShipField(ModuleDefinition module)
    {
        TypeDefinition chunkAreaBiomeSpawnData = module.Types.First(d => d.Name == "ChunkAreaBiomeSpawnData");
        TypeReference tr = module.Import(typeof(byte));
        AddFieldDefinition(chunkAreaBiomeSpawnData, "TownShip", FieldAttributes.Public, tr);
    }
    public void AddHGAbstractField(ModuleDefinition module)
    {
        TypeDefinition hgAbstract = module.Types.First(d => d.Name == "HGAbstract" && d.Namespace == "RWG2.HubGenerators");
        TypeReference tr = module.Import(typeof(int));
        AddFieldDefinition(hgAbstract, "HubChunkStartX", FieldAttributes.Public, tr);
        AddFieldDefinition(hgAbstract, "HubChunkStartY", FieldAttributes.Public, tr);
        AddFieldDefinition(hgAbstract, "HubChunkBitHeight", FieldAttributes.Public, tr);
        AddFieldDefinition(hgAbstract, "HubChunkByteWidth", FieldAttributes.Public, tr);
        tr = module.Import(typeof(byte []));
        AddFieldDefinition(hgAbstract, "HubChunkFlg", FieldAttributes.Public, tr);
    }

    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        classDef.Fields.Add(fd);

    }
    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType, object constantvalue)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        fd.Constant = constantvalue;
        classDef.Fields.Add(fd);

    }
    public void ChangeGameStagePartySpawnerCtor(ModuleDefinition module)
    {
        TypeDefinition gameStagePartySpawner = module.Types.First(d => d.Name == "AIDirectorGameStagePartySpawner");
        MethodDefinition gameStagePartySpawnerCtor = gameStagePartySpawner.Methods.First(d => d.Name == ".ctor");
        Instruction insertPos = null;
        Instruction jumpToPos = null;
        foreach (Instruction inst in gameStagePartySpawnerCtor.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).Name == "KZ")
            {
                jumpToPos = inst.Next;
            }
            if (inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).Name == "GetGameStage")
            {
                insertPos = inst.Previous.Previous;
            }
        }
        ILProcessor ilgameStagePartySpawnerCtor = gameStagePartySpawnerCtor.Body.GetILProcessor();
        ilgameStagePartySpawnerCtor.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_2));
        ilgameStagePartySpawnerCtor.InsertBefore(insertPos, Instruction.Create(OpCodes.Brfalse_S, jumpToPos));

    }

    public bool Patch(ModuleDefinition module)
    {
        Console.WriteLine("==WildernessGameStage Patcher===");
        AddBiomeSpawnEntityGroupDataField(module);
        AddChunkTownShipJAField(module);
        AddChunkDominantTownShipField(module);
        AddChunkAreaBiomeSpawnDataTownShipField(module);
        AddHGAbstractField(module);
        ChangeGameStagePartySpawnerCtor(module);
        RemoveHGSotcketsCZCall(module);
        return true;
    }


}
