﻿using System;
using System.Linq;
using System.Xml;
using System.Collections.Generic;
using UnityEngine;
using System.IO;

public static class WildernessGameStageMethod
{
    public class WildernessGameStageBeforeGroup
    {
        public WildernessGameStageBeforeGroup()
        {


        }
        public WildernessGameStageBeforeGroup(int gameStage, string groupName)
        {
            _gamestage = gameStage;
            _groupName = groupName;

        }
        public void Set(int gameStage, string groupName)
        {
            _gamestage = gameStage;
            _groupName = groupName;

        }
        private int _gamestage;
        public int GameStage
        {
            get
            {
                return _gamestage;
            }
            set
            {
                _gamestage = value;
            }

        }
        private string _groupName;
        public string GroupName
        {
            get
            {
                return _groupName;
            }
            set
            {
                _groupName = value;
            }
        }
    }
    public class WildernessGameStageData
    {
        public WildernessGameStageData()
        {

        }
        public WildernessGameStageData(ulong nextUpdateTime, int gamestage)
        {
            _nextUpdateTime = nextUpdateTime;
            _gamestage = gamestage;
        }
        public void Set(ulong nextUpdateTime, int gamestage)
        {
            _nextUpdateTime = nextUpdateTime;
            _gamestage = gamestage;
        }
        private ulong _nextUpdateTime;
        public ulong NextUpdateTime
        {
            get
            {
                return _nextUpdateTime;
            }
            set
            {
                _nextUpdateTime = value;
            }

        }
        private int _gamestage;
        public int GameStage
        {
            get
            {
                return _gamestage;
            }
            set
            {
                _gamestage = value;
            }

        }
        Dictionary<string, WildernessGameStageBeforeGroup> _beforeGamestageGroup = new Dictionary<string, WildernessGameStageBeforeGroup>();
        public Dictionary<string, WildernessGameStageBeforeGroup> BeforeGamestageGroup
        {
            get
            {
                return _beforeGamestageGroup;
            }
            set
            {
                _beforeGamestageGroup = value;
            }
        }
    }

    public static void InitWildernessGameStageDic()
    {
        _chunkGameStageDic = new Dictionary<long, WildernessGameStageData>();
        _chunkGameStageGroupDic = new Dictionary<int, Dictionary<int, WildernessGameStageData>>();
        _biomeSpawnEntityGroupListDic = new Dictionary<byte, Dictionary<byte, BiomeSpawnEntityGroupList>>();
    }
    private static Dictionary<long, WildernessGameStageData> _chunkGameStageDic = new Dictionary<long, WildernessGameStageData>();
    private static Dictionary<int,Dictionary<int,WildernessGameStageData>> _chunkGameStageGroupDic = new Dictionary<int, Dictionary<int, WildernessGameStageData>>();
    private static uint _gamestageUpdateWaitTime = 500;
    private static int _hubAreaSurplus = 16;
    private static int _gamestageGroupChunk = 16;
    public static void SetGamestageUpdateWaitTime(XmlElement element)
    {
        if (element.HasAttribute("GamestageUpdateWaitTime"))
        {
            _gamestageUpdateWaitTime = uint.Parse(element.GetAttribute("GamestageUpdateWaitTime"));
        }
        else
        {
            _gamestageUpdateWaitTime = 500;
        }
        if (element.HasAttribute("HubAreaSurplus"))
        {
            _hubAreaSurplus = int.Parse(element.GetAttribute("HubAreaSurplus"));
        }
        else
        {
            _hubAreaSurplus = 16;
        }
        if (element.HasAttribute("GamestageGroupChunk"))
        {
            _gamestageGroupChunk = int.Parse(element.GetAttribute("GamestageGroupChunk"));
        }
        else
        {
            _gamestageGroupChunk = 16;
        }

    }
    private static int _wanderingHordeStageMax = 50;//wandering horde stage max
    public static void SetWanderingHordeStageMax(XmlElement element)
    {
        if (element.HasAttribute("WanderingHordeStageMax"))
        {
            _wanderingHordeStageMax = int.Parse(element.GetAttribute("WanderingHordeStageMax"));
        }
        else
        {
            _wanderingHordeStageMax = 50;
        }

    }
    public static int GetWanderingHordeGameStage(int stage,int mod)
    {
        if(_wanderingHordeStageMax > stage)
        {
            return stage;
        }
        else
        {
            int ret = (stage - (_wanderingHordeStageMax - mod)) % mod + (_wanderingHordeStageMax - mod);
            return ret;
        }
    }
    public static void UnLoadChunk(Chunk chunk)
    {
        if (_chunkGameStageDic.ContainsKey(chunk.Key))
        {
            
            _chunkGameStageDic.Remove(chunk.Key);
        }
    }
    public static string GetWildernessGameStageGroup( string spawnerName, ChunkAreaBiomeSpawnData biomeSpawnData, World world)
    {
        int groupX = biomeSpawnData.chunk.X / _gamestageGroupChunk;
        int groupZ = biomeSpawnData.chunk.Z / _gamestageGroupChunk;

        if (!_chunkGameStageGroupDic.ContainsKey(groupX))
        {
            _chunkGameStageGroupDic[groupX] = new Dictionary<int, WildernessGameStageData>();
        }
        if (!_chunkGameStageGroupDic[groupX].ContainsKey(biomeSpawnData.chunk.Z))
        {
            _chunkGameStageGroupDic[groupX][groupZ] = new WildernessGameStageData();
        }
        if (!_chunkGameStageDic.ContainsKey(biomeSpawnData.chunk.Key))
        {
            _chunkGameStageDic[biomeSpawnData.chunk.Key] = _chunkGameStageGroupDic[groupX][groupZ];
        }
        WildernessGameStageData data = _chunkGameStageDic[biomeSpawnData.chunk.Key];
        if (world.worldTime >= data.NextUpdateTime)
        {
            data.NextUpdateTime = world.worldTime + _gamestageUpdateWaitTime;

            AIDirectorGameStagePartySpawner gameStagePartySpawner = new AIDirectorGameStagePartySpawner(world, null);
            List<global::EntityPlayer> players = world.GetPlayers();
            for (int i = 0; i < players.Count; i++)
            {
                if (!players[i].Spawned)
                {
                }
                else
                {
                    Vector3 position = players[i].GetPosition();
                    Rect rect = new Rect(position.x - 40f, position.z - 40f, 80f, 80f);
                    if (rect.Overlaps(biomeSpawnData.area))
                    {
                        gameStagePartySpawner.AddMember(players[i]);
                    }
                }
            }
            gameStagePartySpawner.ResetPartyLevel(0);
            data.GameStage = gameStagePartySpawner.partyLevel;

        }
        else
        {
        }
        bool isNew = !data.BeforeGamestageGroup.ContainsKey(spawnerName);
        if (isNew)
        {
            data.BeforeGamestageGroup[spawnerName] = new WildernessGameStageBeforeGroup();
        }
        if(isNew || data.BeforeGamestageGroup[spawnerName].GameStage != data.GameStage)
        {
            GameStageDefinition spawner = GameStageDefinition.GetGameStage(spawnerName);
            GameStageDefinition.Stage stage = spawner.GetStage(data.GameStage);
            GameStageDefinition.Spawn spawn = stage.GetSpawn(0);
            data.BeforeGamestageGroup[spawnerName].Set(data.GameStage, spawn.groupName);            
        }
        return data.BeforeGamestageGroup[spawnerName].GroupName;
    }
    private static Dictionary<byte, Dictionary<byte, BiomeSpawnEntityGroupList>> _biomeSpawnEntityGroupListDic = new Dictionary<byte, Dictionary<byte, BiomeSpawnEntityGroupList>>();
    public static BiomeSpawnEntityGroupList GetBiomeSpawnEntityGroupList(ChunkAreaBiomeSpawnData chunkAreaBiomeSpawnData, World world)
    {
        byte biomeID = chunkAreaBiomeSpawnData.biomeId;
        byte townShipID = chunkAreaBiomeSpawnData.TownShip;
        //HubCell.TownshipType.None is wilderness
        if (townShipID == (byte)HubCell.TownshipType.None)
        {
            townShipID = (byte)HubCell.TownshipType.Wilderness;
        }
        if (!_biomeSpawnEntityGroupListDic.ContainsKey(biomeID))
        {
            _biomeSpawnEntityGroupListDic.Add(biomeID, new Dictionary<byte, BiomeSpawnEntityGroupList>());
        }
        if(!_biomeSpawnEntityGroupListDic[biomeID].ContainsKey(townShipID))
        {
            global::BiomeDefinition biome = world.Biomes.GetBiome(biomeID);
            BiomeSpawnEntityGroupList biomeSpawnEntityGroupList = global::BiomeSpawningClass.list[biome.m_sBiomeName];
            BiomeSpawnEntityGroupList add = new BiomeSpawnEntityGroupList();
            if (biomeSpawnEntityGroupList != null)
            {
                foreach (BiomeSpawnEntityGroupData data in biomeSpawnEntityGroupList.list)
                {
                    if(data.TownShip.Count == 0 || data.TownShip.Contains((byte)HubCell.TownshipType.Any) || data.TownShip.Contains((byte)HubCell.TownshipType.None) || data.TownShip.Contains(townShipID))
                    {
                        add.list.Add(data);
                    }
                }
            }
            _biomeSpawnEntityGroupListDic[biomeID].Add(townShipID, add);
        }
        BiomeSpawnEntityGroupList ret = _biomeSpawnEntityGroupListDic[biomeID][townShipID];
        if (ret.list.Count == 0)
        {
            return null;
        }
        else
        {
            return ret;
        }
    }
    public static void SetTownship(BiomeSpawnEntityGroupData biomeSpawnEntityGroupData,XmlElement xmlElement)
    {
        biomeSpawnEntityGroupData.TownShip = new List<byte>();
        if(xmlElement.HasAttribute("AllowedTownships"))
        {
            string [] array = xmlElement.GetAttribute("AllowedTownships").Split(new char[] { ',', ' ' }, StringSplitOptions.RemoveEmptyEntries);
            foreach(string townshipst in array)
            {
                if (townshipst.ToLower() == HubCell.TownshipType.Any.ToString().ToLower())
                {
                    biomeSpawnEntityGroupData.TownShip.Add((byte)HubCell.TownshipType.Any);
                }
                else if (townshipst.ToLower() == HubCell.TownshipType.Wilderness.ToString().ToLower())
                {
                    biomeSpawnEntityGroupData.TownShip.Add((byte)HubCell.TownshipType.Wilderness);
                }
                else if (townshipst.ToLower() == HubCell.TownshipType.Rural.ToString().ToLower())
                {
                    biomeSpawnEntityGroupData.TownShip.Add((byte)HubCell.TownshipType.Rural);
                }
                else if (townshipst.ToLower() == HubCell.TownshipType.Town.ToString().ToLower())
                {
                    biomeSpawnEntityGroupData.TownShip.Add((byte)HubCell.TownshipType.Town);
                }
                else if (townshipst.ToLower() == HubCell.TownshipType.City.ToString().ToLower())
                {
                    biomeSpawnEntityGroupData.TownShip.Add((byte)HubCell.TownshipType.City);
                }
                else
                {
                    biomeSpawnEntityGroupData.TownShip.Add((byte)HubCell.TownshipType.None);
                }
            }
        }
    }
    public static void CalcDominantTownShip(Chunk chunk)
    {
        int[] array = new int[50];
        for (int i = 0; i < chunk.TownShipJA.Length; i++)
        {
            array[(int)chunk.TownShipJA[i]]++;
        }
        int num = 0;
        for (int j = 0; j < array.Length; j++)
        {
            if (array[j] > num)
            {
                chunk.DominantTownShip = (byte)j;
                num = array[j];
            }
        }

    }
    public static void CheckAndSetTownShipID(int x, int z, int chunkX,int chunkZ,RWG2.HubCell hub, Chunk chunk)
    {
        if(hub.HubGenerator != null && hub.HubGenerator.BoundingBox.Contains(new Vector2(x,z)))
        {
            int EndX = hub.HubGenerator.HubChunkStartX + hub.HubGenerator.HubChunkByteWidth * 8;
            int EndY = hub.HubGenerator.HubChunkStartY + hub.HubGenerator.HubChunkBitHeight;
            if(hub.HubGenerator.HubChunkStartX <= chunk.X && EndX > chunk.X
                && hub.HubGenerator.HubChunkStartY <= chunk.Z && EndY > chunk.Z)
            {
                int hubChunkX = chunk.X - hub.HubGenerator.HubChunkStartX;
                int hubChunkY = chunk.Z - hub.HubGenerator.HubChunkStartY;
                if ((hub.HubGenerator.HubChunkFlg[hubChunkX / 8 + hubChunkY * hub.HubGenerator.HubChunkByteWidth] & (0x01 << hubChunkX % 8)) != 0)
                {
                    SetTownShipID(chunkX, chunkZ, (byte)hub.HubGenerator.township, chunk);
                }
            }
        }
    }
    public static void SetTownShipID(int x,int z,byte townshipID,Chunk chunk)
    {
        chunk.TownShipJA[x + z * 16] = townshipID;

    }
    public static string GetHubChunkFlgFileName(Vector2i grid)
    {
        return global::GameUtils.GetSaveGameDir() + string.Format("/HubCellData/{0}.{1}.hcf",grid.x,grid.y);
    }
    public static void WriteHubChunkFlgData(RWG2.HubGenerators.HGAbstract hgAbst)
    {
        if (!global::Steam.Network.IsClient)
        {
            string fileName = GetHubChunkFlgFileName(hgAbst.hubCell.GridPos);
            BinaryWriter bw = new BinaryWriter(File.OpenWrite(fileName));

            bw.Write(hgAbst.HubChunkStartX);
            bw.Write(hgAbst.HubChunkStartY);
            bw.Write(hgAbst.HubChunkBitHeight);
            bw.Write(hgAbst.HubChunkByteWidth);
            if (hgAbst.HubChunkBitHeight != 0 && hgAbst.HubChunkByteWidth != 0)
            {
                bw.Write(hgAbst.HubChunkFlg, 0, hgAbst.HubChunkByteWidth * hgAbst.HubChunkBitHeight);
            }
            bw.Close();

        }
    }

    public static void ReadHubChunkFlgData(RWG2.HubCell hubCell)
    {
        if (!global::Steam.Network.IsClient)
        {
            RWG2.HubGenerators.HGAbstract hgAbst = hubCell.HubGenerator;
            string fileName = GetHubChunkFlgFileName(hubCell.GridPos);
            BinaryReader br = new BinaryReader(File.OpenRead(fileName));

            hgAbst.HubChunkStartX = br.ReadInt32();
            hgAbst.HubChunkStartY = br.ReadInt32();
            hgAbst.HubChunkBitHeight = br.ReadInt32();
            hgAbst.HubChunkByteWidth = br.ReadInt32();
            if (hgAbst.HubChunkBitHeight != 0 && hgAbst.HubChunkByteWidth != 0)
            {
                hgAbst.HubChunkFlg = br.ReadBytes(hgAbst.HubChunkByteWidth * hgAbst.HubChunkBitHeight);
            }
            br.Close();
        }
    }
    public static void SetAreaHubDataLot(RWG2.HubGenerators.HGAbstract hgAbst,RWG2.HubCell.Lot lot)
    {
        SetAreaHubData(hgAbst, lot.Area);

    }
    public static void SetAreaHubData(RWG2.HubGenerators.HGAbstract hgAbst,UnityEngine.Rect boundingBox )
    {
        int EndX = hgAbst.HubChunkStartX + hgAbst.HubChunkByteWidth * 8;
        int EndY = hgAbst.HubChunkStartY + hgAbst.HubChunkBitHeight;
        float bl = boundingBox.position.x - _hubAreaSurplus;
        float bt = boundingBox.position.y - _hubAreaSurplus;
        float br = bl + boundingBox.size.x + _hubAreaSurplus * 2;
        float bb = bt + boundingBox.size.y + _hubAreaSurplus * 2;
        int left = (int)(bl);
        if (bl < left)
        {
            left--;
        }
        int top = (int)(bt);
        if (bt < top)
        {
            top--;
        }
        int right = (int)(br);
        if (br > right)
        {
            right++;
        }
        int bottom = (int)(bb);
        if (bb > bottom)
        {
            bottom++;
        }
        int chunkLeft = left / 16;
        if (left % 16 < 0)
        {
            chunkLeft--;
        }
        int chunkTop = top / 16;
        if (top % 16 < 0)
        {
            chunkTop--;
        }
        int chunkRight = right / 16;
        if (right % 16 > 0)
        {
            chunkRight++;
        }
        int chunkBottom = bottom / 16;
        if (bottom % 16 > 0)
        {
            chunkBottom++;
        }
        for (int i = chunkLeft; i <= chunkRight; i++)
        {
            if (hgAbst.HubChunkStartX <= i && EndX > i)
            {
                for (int j = chunkTop; j <= chunkBottom; j++)
                {
                    if (hgAbst.HubChunkStartY <= j && EndY > j)
                    {
                        int x = i - hgAbst.HubChunkStartX;
                        int y = j - hgAbst.HubChunkStartY;
                        hgAbst.HubChunkFlg[x / 8 + y * hgAbst.HubChunkByteWidth] |= (byte)(0x01 << (x % 8));
                    }
                }
            }
        }

    }
    public static void InitHGAbstructHubData(RWG2.HubCell hub)
    {
        RWG2.HubGenerators.HGAbstract hgAbst = hub.HubGenerator;
        hgAbst.HubChunkStartX = (int)(hgAbst.BoundingBox.position.x) / 16;
        if ((int)(hgAbst.BoundingBox.position.x) < 0 && (int)(hgAbst.BoundingBox.position.x) % 16 != 0)
        {
            hgAbst.HubChunkStartX--;
        }
        hgAbst.HubChunkStartY = (int)(hgAbst.BoundingBox.position.y) / 16;
        if ((int)(hgAbst.BoundingBox.position.y) < 0 && (int)(hgAbst.BoundingBox.position.y) % 16 != 0)
        {
            hgAbst.HubChunkStartY--;
        }
        int HubChunkOffsetX = hgAbst.HubChunkStartX * 16;
        int HubChunkOffsetY = hgAbst.HubChunkStartY * 16;
        int HubChunkEndX = (HubChunkOffsetX + (int)(hgAbst.BoundingBox.size.x)) / 16;
        if((HubChunkOffsetX + (int)(hgAbst.BoundingBox.size.x)) % 16 != 0)
        {
            HubChunkEndX++;
        }
        int HubChunkBitWidth = HubChunkEndX - hgAbst.HubChunkStartX;
        int hubChunkEndY = (HubChunkOffsetY + (int)(hgAbst.BoundingBox.size.y)) / 16;
        if ((HubChunkOffsetY + (int)(hgAbst.BoundingBox.size.y)) % 16 != 0)
        {
            hubChunkEndY++;
        }
        hgAbst.HubChunkBitHeight = hubChunkEndY - hgAbst.HubChunkStartY + 1;
        hgAbst.HubChunkByteWidth = (HubChunkBitWidth / 8) +1;
        if (HubChunkBitWidth % 8 != 0)
        {
            hgAbst.HubChunkByteWidth++;
        }
        
        hgAbst.HubChunkFlg = new byte[hgAbst.HubChunkByteWidth * hgAbst.HubChunkBitHeight];
        List<RWG2.PathGenerators.PathAbstract> list = hgAbst.GetPaths();
        foreach (RWG2.PathGenerators.PathAbstract abst in list)
        {
            SetAreaHubData(hgAbst, abst.BoundingBox);
        }
    }
}
